/*
 * Part of Very Secure FTPd
 * Licence: GPL v2
 * Author: Dmitriy Balashov
 * http_str.c
 *
 * HTTP messages.
 */

#include "defs.h"
#include "http_str.h"
#include "str.h"
#include "strlist.h"
#include "sysutil.h"
#include "tunables.h"
#include "charconv.h"
#include "vsftpver.h"
#include "session.h"
#include "utility.h"

#define digit_to_int(code) ((code < 'A') ? (code - '0') : ((code & ~0x20) - 'A' + 10))
#define digits_to_int(code1, code2) ((digit_to_int(code1) << 4) | digit_to_int(code2))
#define is_digit(code) ((code >= '0' && code <= '9') || ((code & ~0x20) >= 'A' && (code & ~0x20) <= 'F'))

#define int_to_digit(code) ("0123456789abcdef"[code] + 0)

static const unsigned char urlchr_table[256] =
{
  1,  1,  1,  1,   1,  1,  1,  1,   /* NUL SOH STX ETX  EOT ENQ ACK BEL */
  1,  1,  1,  1,   1,  1,  1,  1,   /* BS  HT  LF  VT   FF  CR  SO  SI  */
  1,  1,  1,  1,   1,  1,  1,  1,   /* DLE DC1 DC2 DC3  DC4 NAK SYN ETB */
  1,  1,  1,  1,   1,  1,  1,  1,   /* CAN EM  SUB ESC  FS  GS  RS  US  */
  1,  0,  1,  1,   0,  1,  0,  0,   /* SP  !   "   #    $   %   &   '   */
  0,  0,  0,  0,   0,  0,  0,  0,   /* (   )   *   +    ,   -   .   /   */
  0,  0,  0,  0,   0,  0,  0,  0,   /* 0   1   2   3    4   5   6   7   */
  0,  0,  1,  0,   1,  0,  1,  0,   /* 8   9   :   ;    <   =   >   ?   */
  1,  0,  0,  0,   0,  0,  0,  0,   /* @   A   B   C    D   E   F   G   */
  0,  0,  0,  0,   0,  0,  0,  0,   /* H   I   J   K    L   M   N   O   */
  0,  0,  0,  0,   0,  0,  0,  0,   /* P   Q   R   S    T   U   V   W   */
  0,  0,  0,  1,   1,  1,  1,  0,   /* X   Y   Z   [    \   ]   ^   _   */
  1,  0,  0,  0,   0,  0,  0,  0,   /* `   a   b   c    d   e   f   g   */
  0,  0,  0,  0,   0,  0,  0,  0,   /* h   i   j   k    l   m   n   o   */
  0,  0,  0,  0,   0,  0,  0,  0,   /* p   q   r   s    t   u   v   w   */
  0,  0,  0,  1,   1,  1,  0,  1,   /* x   y   z   {    |   }   ~   DEL */

  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,
  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,
  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,
  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,

  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,
  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,
  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,
  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1
};

void
str_escape(struct mystr* p_str)
{
  const char* srcbuf;
  unsigned int srclen;
  char* dstbuf;
  unsigned int srcpos = 0, dstpos = 0;
  unsigned int char_code;

  srclen = str_getlen(p_str);     // Len of source string
  srcbuf = str_getbuf(p_str);
  dstbuf = vsf_sysutil_malloc(srclen * 3);

  while (srcpos < srclen)
  {
    char_code = (unsigned char)srcbuf [srcpos++];

    if (urlchr_table[char_code])
    {
      dstbuf [dstpos++] = '%';
      dstbuf [dstpos++] = int_to_digit(char_code >> 4);
      dstbuf [dstpos++] = int_to_digit(char_code & 0x0f);
    }
    else
    {
      dstbuf [dstpos++] = char_code;
    }
  }

  dstbuf [dstpos] = '\0';

  str_alloc_text(p_str, dstbuf);
  vsf_sysutil_free(dstbuf);
}

void
str_unescape(struct mystr* p_str)
{
  const char* srcbuf;
  unsigned int srclen;
  char* dstbuf;
  unsigned int srcpos = 0, dstpos = 0;
  unsigned int char_code, char2_code;

  srclen = str_getlen(p_str);     // Len of source string
  srcbuf = str_getbuf(p_str);
  dstbuf = vsf_sysutil_malloc(srclen+1);

  while (srcpos < srclen)
  {
    char_code = (unsigned char)srcbuf [srcpos++];
    if (char_code == '%' && srcpos < (srclen - 2) &&
        is_digit(srcbuf [srcpos]) && is_digit(srcbuf [srcpos+1]))
    {
      char_code = srcbuf [srcpos++];
      char2_code = srcbuf [srcpos++];
      dstbuf [dstpos++] = digits_to_int (char_code, char2_code);
    }
    else
    {
      dstbuf [dstpos++] = char_code;
    }
  }

  dstbuf [dstpos] = '\0';

  str_alloc_text(p_str, dstbuf);
  vsf_sysutil_free(dstbuf);
}

void
str_process_template(struct vsf_session* p_sess,
                     struct mystr* p_str,
                     struct mystr* p_tpl,
                     const struct mystr* p_filename,
                     const struct vsf_sysutil_statbuf* p_stat,
                     long curr_time,
                     struct mystr_list* p_dir)
{
  static struct mystr s_lhs_chunk_str;
  static struct mystr s_rhs_chunk_str;
  char cmd;
  filesize_t size;
  long f_time;

  if (p_stat)
  {
    f_time = vsf_sysutil_statbuf_get_filedate(p_stat);
    size = vsf_sysutil_statbuf_get_size(p_stat);
  }
  else
  {
    f_time = curr_time;
    size = 0;
  }

  str_free(p_str);
  str_reserve(p_str, VSFTP_DIR_BUFSIZE);
  str_copy(&s_lhs_chunk_str, p_tpl);

  while (str_getlen(&s_lhs_chunk_str) > 0)
  {
    str_split_char(&s_lhs_chunk_str, &s_rhs_chunk_str, '%');
    str_append_str(p_str, &s_lhs_chunk_str);
    if (str_getlen(&s_rhs_chunk_str) > 0)
    {
      cmd = str_del_char_at(&s_rhs_chunk_str, 0);

      if (p_filename && str_getlen(p_filename) > 0)
      {
        static struct mystr p_tmp_str = INIT_MYSTR;
        switch (cmd)
        {
          case 'i':
          case 'I':
            str_append_str(p_str, p_filename);
            break;
          case 'd': // Date
            str_append_text(p_str, vsf_sysutil_get_sdate(tunable_use_localtime, f_time));
            break;
          case 'D': // Full date time in format
            str_append_text(p_str, vsf_sysutil_get_datetime(tunable_use_localtime, f_time));
            break;
          case 't': // Time
            str_append_text(p_str, vsf_sysutil_get_time(tunable_use_localtime, f_time));
            break;
          case 'T': // Date time if FTP format
            str_append_text(p_str, vsf_sysutil_statbuf_get_date(p_stat,
                                                      tunable_use_localtime,
                                                      curr_time));
            break;
          case 'p':
          case 'P': // Permissions as in FTP
            str_append_text(p_str, vsf_sysutil_statbuf_get_perms(p_stat));
            break;
          case 'f':
          case 'F': // File/Dir/Link
            if (vsf_sysutil_statbuf_is_dir(p_stat))
            {
              str_append_text(p_str, "dir");
            }
            else
            if (vsf_sysutil_statbuf_is_symlink(p_stat))
            {
              str_append_text(p_str, "link");
            }
            else
            if (vsf_sysutil_statbuf_is_regfile(p_stat))
            {
              str_append_text(p_str, "file");
            }
            else
            if (vsf_sysutil_statbuf_is_socket(p_stat))
            {
              str_append_text(p_str, "sock");
            }
            else
            {
              str_append_text(p_str, "unk");
            }
            break;
          case 'l':
          case 'L': // Link?
            str_copy(&p_tmp_str, p_filename);
            str_escape(&p_tmp_str);
            str_append_str(p_str, &p_tmp_str);
            if (vsf_sysutil_statbuf_is_dir(p_stat))
            {
              str_append_text(p_str, "/");
            }
            break;
          case 's': // File size in bytes
            if (vsf_sysutil_statbuf_is_regfile(p_stat))
            {
              str_append_filesize_t(p_str, size);
            }
            else
            {
              str_append_text(p_str, "&nbsp;");
            }
            break;
          case 'S': // File size in Kbytes
            if (vsf_sysutil_statbuf_is_regfile(p_stat))
            {
              str_append_filesize_t(p_str, (size+512)/1024);
              str_append_char(p_str, 'K');
            }
            else
            {
              str_append_text(p_str, "&nbsp;");
            }
            break;
          case 'u':
          case 'U': // User
            str_getuser(&p_tmp_str, p_stat);
            str_append_str(p_str, &p_tmp_str);
            break;
          case 'g':
          case 'G': // Group
            str_getgroup(&p_tmp_str, p_stat);
            str_append_str(p_str, &p_tmp_str);
            break;
          default:
            str_append_char(p_str, '%');
            str_append_char(p_str, cmd);
            break;
        }
      }
      else
      {
        switch (cmd)
        {
          case 'b':
          case 'B':
            if (p_dir)
            {
              unsigned int dir_index_max = str_list_get_length(p_dir);
              unsigned int dir_index;

              for (dir_index = 0; dir_index < dir_index_max; dir_index++)
              {
                str_append_str(p_str, str_list_get_pstr(p_dir, dir_index));
              }
            }
            /// Browse directory
            break;
          case 'r':
          case 'R':
            str_append_str(p_str, &p_sess->http_get_arg);
            break;
          case 'd':
            str_append_text(p_str, vsf_sysutil_get_sdate(tunable_use_localtime, curr_time));
            break;
          case 'D':
            str_append_text(p_str, vsf_sysutil_get_date(tunable_use_localtime, curr_time));
            break;
          case 't':
            str_append_text(p_str, vsf_sysutil_get_time(tunable_use_localtime, curr_time));
            break;
          case 'T':
            str_append_text(p_str, vsf_sysutil_get_datetime(tunable_use_localtime, curr_time));
          break;
          case 'v':
            str_append_text(p_str, "vsFTPd");
            break;
          case 'V':
            str_append_text(p_str, "vsFTPd " VSF_VERSION);
            break;
          default:
            str_append_char(p_str, '%');
            str_append_char(p_str, cmd);
            break;
        }
      }
    }
    str_copy(&s_lhs_chunk_str, &s_rhs_chunk_str);
  }
}

void
str_getuser(struct mystr* p_str, const struct vsf_sysutil_statbuf* p_stat)
{
  if (tunable_hide_ids)
  {
    str_alloc_text(p_str, "ftp");
  }
  else
  {
    int uid = vsf_sysutil_statbuf_get_uid(p_stat);
    struct vsf_sysutil_user* p_user = 0;
    if (tunable_text_userdb_names)
    {
      p_user = vsf_sysutil_getpwuid(uid);
    }
    if (p_user == 0)
    {
      str_alloc_ulong(p_str, (unsigned long) uid);
    }
    else
    {
      str_alloc_text(p_str, vsf_sysutil_user_getname(p_user));
    }
  }
}

void
str_getgroup(struct mystr* p_str, const struct vsf_sysutil_statbuf* p_stat)
{
  if (tunable_hide_ids)
  {
    str_alloc_text(p_str, "ftp");
  }
  else
  {
    int gid = vsf_sysutil_statbuf_get_gid(p_stat);
    struct vsf_sysutil_group* p_group = 0;
    if (tunable_text_userdb_names)
    {
      p_group = vsf_sysutil_getgrgid(gid);
    }
    if (p_group == 0)
    {
      str_alloc_ulong(p_str, (unsigned long) gid);
    }
    else
    {
      str_alloc_text(p_str, vsf_sysutil_group_getname(p_group));
    }
  }
}

