/*
 * Part of Very Secure FTPd
 * Licence: GPL v2
 * Author: Dmitriy Balashov
 * pasvrules.c
 */

#include "pasvrules.h"
#include "tunables.h"
#include "str.h"
#include "session.h"
#include "filestr.h"
#include "defs.h"
#include "sysutil.h"
#include "logging.h"
#include "ftpcodes.h"
#include "ftpcmdio.h"
#include "charconv.h"
#include "utility.h"

static int parse_rules(struct mystr* p_rule_str, pasv_rules rule);
pasv_rules find_rule(struct vsf_session* p_sess);
void dispose_rules();

static pasv_rules rules_list = 0;

void
vsf_pasvrules_load(const char* p_filename)
{
  struct mystr rules_file_str = INIT_MYSTR;
  struct mystr rule_setting_str = INIT_MYSTR;
  unsigned int str_pos = 0;
  int retval;
  pasv_rules rule = 0, owner = 0;

  retval = str_fileread(&rules_file_str, p_filename, VSFTP_CONF_FILE_MAX);
  if (vsf_sysutil_retval_is_error(retval))
  {
    return;
  }

  while (str_getline(&rules_file_str, &rule_setting_str, &str_pos))
  {
    if (str_isempty(&rule_setting_str) ||
        str_get_char_at(&rule_setting_str, 0) == '#')
    {
      continue;
    }
    if (!rules_list)
    {
      rule = vsf_sysutil_malloc_zero(sizeof(pasv_rule));
      rules_list = rule;
    }
    else
    {
      owner = rule;
      rule = vsf_sysutil_malloc_zero(sizeof(pasv_rule));
      owner->next = rule;
    }
    if (parse_rules(&rule_setting_str, rule) == 0)
    {
      vsf_sysutil_free(rule);
      if (owner)
      {
        owner->next = 0;
      }
      else
      {
        rules_list = 0;
      }
    }
  }
  str_free(&rules_file_str);
  if (!str_isempty(&rule_setting_str))
  {
    str_free(&rule_setting_str);
  }
  if (rules_list && tunable_add_default_rule)
  {
    owner = rule;
    rule = vsf_sysutil_malloc_zero(sizeof(pasv_rule));
    owner->next = rule;

    rule->remote_charset  = vsf_charconv_codepage(tunable_remote_charset);
    rule->allow_anonymous = tunable_anonymous_enable;
    rule->anon_upl_enable = tunable_anon_upload_enable;
    rule->anon_md_enable  = tunable_anon_mkdir_write_enable;
    rule->anon_oth_enable = tunable_anon_other_write_enable;
    rule->anon_del_enable = tunable_anon_delete_enable;
  }
}

void
vsf_pasvrules_define(struct vsf_session* p_sess)
{
  static pasv_rules rules;
  rules = find_rule(p_sess);

  if (rules)
  {
    tunable_anonymous_enable        = rules->allow_anonymous;
    tunable_anon_upload_enable      = rules->anon_upl_enable;
    tunable_anon_mkdir_write_enable = rules->anon_md_enable;
    tunable_anon_other_write_enable = rules->anon_oth_enable;
    tunable_anon_delete_enable      = rules->anon_del_enable;
    if (tunable_pasv_address)
    {
      vsf_sysutil_free((char*) tunable_pasv_address);
    }
    tunable_pasv_address = str_strdup(&rules->pasv_address);

    p_sess->remote_charset  = rules->remote_charset;
    dispose_rules();
  }
  else
  {
    p_sess->remote_charset  = vsf_charconv_codepage(tunable_remote_charset);
  }

  /* Resolve pasv_address if required */
  if (tunable_listen && tunable_pasv_address && tunable_pasv_addr_resolve)
  {
    struct vsf_sysutil_sockaddr* p_addr = 0;
    const char* p_numeric_addr;
    vsf_sysutil_dns_resolve(&p_addr, tunable_pasv_address);
    vsf_sysutil_free((char*) tunable_pasv_address);
    if (vsf_sysutil_sockaddr_is_any(p_addr))
    {
      p_numeric_addr = vsf_sysutil_inet_ntop(p_sess->p_local_addr);
    }
    else
    {
      p_numeric_addr = vsf_sysutil_inet_ntop(p_addr);
    }
    tunable_pasv_address = vsf_sysutil_strdup(p_numeric_addr);
    vsf_sysutil_free(p_addr);
  }
}

static int
parse_rules(struct mystr* p_rule_str, pasv_rules rule)
{
  static struct mystr s_lhs_chunk_str;
  static struct mystr s_rhs_chunk_str;
  int res_bool;
  int is_remote_charset = 0;

  str_replace_char(p_rule_str, '\t', ' ');
  str_trim_char(p_rule_str, ' ');

  str_copy(&s_lhs_chunk_str, p_rule_str);
  str_free(p_rule_str);

  if (tunable_listen_ipv6)
  {
    vsf_sysutil_sockaddr_alloc_ipv6(&rule->local_ip);
    vsf_sysutil_sockaddr_alloc_ipv6(&rule->remote_mask1);
    vsf_sysutil_sockaddr_alloc_ipv6(&rule->remote_mask2);
  }
  else
  {
    vsf_sysutil_sockaddr_alloc_ipv4(&rule->local_ip);
    vsf_sysutil_sockaddr_alloc_ipv4(&rule->remote_mask1);
    vsf_sysutil_sockaddr_alloc_ipv4(&rule->remote_mask2);
  }

  str_split_char(&s_lhs_chunk_str, &s_rhs_chunk_str, ' '); // Get local_ip
  str_append_char(&s_lhs_chunk_str, '\0');
  if (vsf_sysutil_inet_aton(str_getbuf(&s_lhs_chunk_str), rule->local_ip) == 0)
  {
    return 0;
  }

  str_copy(&s_lhs_chunk_str, &s_rhs_chunk_str); // Get remote_mask1
  str_split_char(&s_lhs_chunk_str, &s_rhs_chunk_str, '/');
  str_append_char(&s_lhs_chunk_str, '\0');
  if (vsf_sysutil_inet_aton(str_getbuf(&s_lhs_chunk_str), rule->remote_mask1) == 0)
  {
    return 0;
  }

  str_copy(&s_lhs_chunk_str, &s_rhs_chunk_str); // Get remote_mask2
  str_split_char(&s_lhs_chunk_str, &s_rhs_chunk_str, ' ');
  if (vsf_sysutil_inet_aton(str_getbuf(&s_lhs_chunk_str), rule->remote_mask2) == 0)
  {
    vsf_sysutil_sockaddr_set_mask(rule->remote_mask2, str_atoi(&s_lhs_chunk_str));
  }

  str_copy(&s_lhs_chunk_str, &s_rhs_chunk_str); // Get pasv_address
  str_split_char(&s_lhs_chunk_str, &s_rhs_chunk_str, ' ');

  str_copy(&rule->pasv_address, &s_lhs_chunk_str);
  str_append_char(&rule->pasv_address, '\0');

  rule->remote_charset  = vsf_charconv_codepage(tunable_remote_charset);
  rule->allow_anonymous = tunable_anonymous_enable;
  rule->anon_upl_enable = tunable_anon_upload_enable;
  rule->anon_md_enable  = tunable_anon_mkdir_write_enable;
  rule->anon_oth_enable = tunable_anon_other_write_enable;
  rule->anon_del_enable = tunable_anon_delete_enable;

  if (!str_isempty(&s_rhs_chunk_str))
  {  // Get remote_charset
    str_copy(&s_lhs_chunk_str, &s_rhs_chunk_str);
    str_split_char(&s_lhs_chunk_str, &s_rhs_chunk_str, ' ');
    str_upper(&s_lhs_chunk_str);
    if (str_equal_bool(&s_lhs_chunk_str) >= 0)
    {
      is_remote_charset = 0;
    }
    else
    {
      str_append_char(&s_lhs_chunk_str, '\0');
      rule->remote_charset = vsf_charconv_codepage(str_getbuf(&s_lhs_chunk_str));
      is_remote_charset = 1;
    }
  }
  if (!str_isempty(&s_rhs_chunk_str) || !is_remote_charset)
  {  // Get allow_anonymous
    if (is_remote_charset)
    {
      str_copy(&s_lhs_chunk_str, &s_rhs_chunk_str);
      str_split_char(&s_lhs_chunk_str, &s_rhs_chunk_str, ' ');
    }
    if ((res_bool = str_equal_bool(&s_lhs_chunk_str)) >= 0)
    {
      rule->allow_anonymous = res_bool;
    }
  }

  if (!str_isempty(&s_rhs_chunk_str))
  {  // Get anon_upl_enable
    str_copy(&s_lhs_chunk_str, &s_rhs_chunk_str);
    str_split_char(&s_lhs_chunk_str, &s_rhs_chunk_str, ' ');
    if ((res_bool = str_equal_bool(&s_lhs_chunk_str)) >= 0)
    {
      rule->anon_upl_enable = res_bool;
    }
  }

  if (!str_isempty(&s_rhs_chunk_str))
  {  // Get anon_md_enable
    str_copy(&s_lhs_chunk_str, &s_rhs_chunk_str);
    str_split_char(&s_lhs_chunk_str, &s_rhs_chunk_str, ' ');
    if ((res_bool = str_equal_bool(&s_lhs_chunk_str)) >= 0)
    {
      rule->anon_md_enable = res_bool;
    }
  }

  if (!str_isempty(&s_rhs_chunk_str))
  {  // Get anon_oth_enable
    str_copy(&s_lhs_chunk_str, &s_rhs_chunk_str);
    str_split_char(&s_lhs_chunk_str, &s_rhs_chunk_str, ' ');
    if ((res_bool = str_equal_bool(&s_lhs_chunk_str)) >= 0)
    {
      rule->anon_oth_enable = res_bool;
    }
  }

  if (!str_isempty(&s_rhs_chunk_str))
  {  // Get anon_del_enable
    str_copy(&s_lhs_chunk_str, &s_rhs_chunk_str);
    str_split_char(&s_lhs_chunk_str, &s_rhs_chunk_str, ' ');
    if ((res_bool = str_equal_bool(&s_lhs_chunk_str)) >= 0)
    {
      rule->anon_del_enable = res_bool;
    }
  }

  return 1;
}

void
dispose_rules()
{
  static pasv_rules rules;

  while (rules_list)
  {
    rules = rules_list->next;
    vsf_sysutil_sockaddr_clear(&rules_list->local_ip);
    vsf_sysutil_sockaddr_clear(&rules_list->remote_mask1);
    vsf_sysutil_sockaddr_clear(&rules_list->remote_mask2);
    str_free(&rules_list->pasv_address);
    vsf_sysutil_free(rules_list);
    rules_list = rules;
  }
}

pasv_rules
find_rule(struct vsf_session* p_sess)
{
  static pasv_rules rules;
  rules = rules_list;

  while (rules)
  {
    if ((vsf_sysutil_sockaddr_is_any(rules->local_ip) ||
         vsf_sysutil_sockaddr_addr_equal(rules->local_ip, p_sess->p_local_addr)) &&
         vsf_sysutil_sockaddr_in_mask(p_sess->p_remote_addr,
                                      rules->remote_mask1,
                                      rules->remote_mask2))
    {
      break;
    }
    rules = rules->next;
  }

  return rules;
}
