/*
 * Part of Very Secure FTPd
 * Licence: GPL v2
 * Author: Dmitriy Balashov
 * usersip.c
 */

#include "usersip.h"
#include "tunables.h"
#include "str.h"
#include "filestr.h"
#include "sysutil.h"
#include "session.h"
#include "defs.h"
#include "logging.h"
#include "utility.h"
#include "charconv.h"

static void parse_rules(struct mystr* p_rule_str);

static ip_rules rules_list = 0;

void
vsf_userip_load(const char* p_filename)
{
  struct mystr rules_file_str = INIT_MYSTR;
  struct mystr rule_setting_str = INIT_MYSTR;
  int retval;
  unsigned int str_pos = 0;

  retval = str_fileread(&rules_file_str, p_filename, VSFTP_CONF_FILE_MAX);
  if (vsf_sysutil_retval_is_error(retval))
  {
    return;
  }

  while (str_getline(&rules_file_str, &rule_setting_str, &str_pos))
  {
    if (str_isempty(&rule_setting_str) ||
        str_get_char_at(&rule_setting_str, 0) == '#')
    {
      continue;
    }
    parse_rules(&rule_setting_str);
  }

  str_free(&rules_file_str);
  if (!str_isempty(&rule_setting_str))
  {
    str_free(&rule_setting_str);
  }

  if (rules_list) {
    while (rules_list->prev)
    {
      rules_list = rules_list->prev;
    }
  }
}

int
vsf_userip_check(const struct mystr* p_user,
                 struct vsf_session* p_sess)
{
  int accept_login = (tunable_users_access_ip_deny == 1) ? 0 : 1;
  static ip_rules rules;

  if (!rules_list)
  {
    return(1);
  }

  rules = rules_list;

  do
  {
    if (str_equal(p_user, &rules->p_user))
    {
      accept_login = 0;
      if (vsf_sysutil_sockaddr_in_mask(p_sess->p_remote_addr,
                                       rules->remote_ip,
                                       rules->remote_mask))
      {
        p_sess->remote_charset = rules->codepage;
        return (1);
      }
    }
    rules = rules->next;
  }
  while (rules);

  return (accept_login);
}

void
vsf_userip_free()
{
  if (rules_list)
  {
    do
    {
      if (rules_list->next)
      {
        rules_list = rules_list->next;
        vsf_sysutil_sockaddr_clear(&rules_list->remote_ip);
        vsf_sysutil_sockaddr_clear(&rules_list->remote_mask);
        vsf_sysutil_free (rules_list->prev);
      } else
      {
        vsf_sysutil_sockaddr_clear(&rules_list->remote_ip);
        vsf_sysutil_sockaddr_clear(&rules_list->remote_mask);
        vsf_sysutil_free (rules_list);
        rules_list = 0;
      }
    }
    while (rules_list);
  }
}

static void
parse_rules(struct mystr* p_rule_str)
{
  static struct mystr s_lhs_chunk_str;
  static struct mystr s_rhs_chunk_str;
  static struct mystr login;
  static struct str_locate_result locate;
  static struct mystr s_ip;
  static struct vsf_sysutil_sockaddr *addr_remote = 0;
  static struct vsf_sysutil_sockaddr *addr_mask;
  int codepage = 0;

  str_replace_char(p_rule_str, '\t', ' ');
  str_trim_char(p_rule_str, ' ');

  str_copy(&s_lhs_chunk_str, p_rule_str);
  str_split_char(&s_lhs_chunk_str, &s_rhs_chunk_str, ' '); // Get user login
  str_copy(&login, &s_lhs_chunk_str);

  str_copy(&s_lhs_chunk_str, &s_rhs_chunk_str);
  str_split_char(&s_lhs_chunk_str, &s_rhs_chunk_str, ' '); // Get user codepage
  codepage = vsf_charconv_codepage(str_getbuf(&s_lhs_chunk_str));
  if (codepage == 0)
  {
    codepage = vsf_charconv_codepage(tunable_remote_charset);
    str_append_char(&s_lhs_chunk_str, ' ');
    str_append_str(&s_lhs_chunk_str, &s_rhs_chunk_str);
    str_copy(&s_rhs_chunk_str, &s_lhs_chunk_str);
  }

  do
  {
    if (!addr_remote)
    {
      if (tunable_listen_ipv6)
      {
        vsf_sysutil_sockaddr_alloc_ipv6(&addr_mask);
        vsf_sysutil_sockaddr_alloc_ipv6(&addr_remote);
      }
      else
      {
        vsf_sysutil_sockaddr_alloc_ipv4(&addr_mask);
        vsf_sysutil_sockaddr_alloc_ipv4(&addr_remote);
      }
    }

    str_copy(&s_lhs_chunk_str, &s_rhs_chunk_str);
    str_split_char(&s_lhs_chunk_str, &s_rhs_chunk_str, ' ');

    locate = str_locate_char(&s_lhs_chunk_str, '/');
    if (locate.found == 1)
    {
      str_copy(&s_ip, &s_lhs_chunk_str);
      str_split_char(&s_ip, &s_lhs_chunk_str, '/');
      str_append_char(&s_ip, '\0');
      if (vsf_sysutil_inet_aton(str_getbuf(&s_ip), addr_remote) != 0)
      {
        if (vsf_sysutil_inet_aton(str_getbuf(&s_lhs_chunk_str), addr_mask) == 0)
        {
          vsf_sysutil_sockaddr_set_mask(addr_mask, str_atoi(&s_lhs_chunk_str));
        }
      }
      else
      {
        continue;
      }
    }
    else
    {
      str_append_char(&s_lhs_chunk_str, '\0');
      if (vsf_sysutil_inet_aton(str_getbuf(&s_lhs_chunk_str), addr_remote) != 0)
      {
        vsf_sysutil_sockaddr_set_mask(addr_mask, ((tunable_listen_ipv6) ? 128 : 32));
      }
      else
      {
        continue;
      }
    }

    if (!rules_list)
    {
      rules_list = vsf_sysutil_malloc(sizeof(ip_rule));
      vsf_sysutil_fillbuff(rules_list, sizeof(ip_rule), 0);
    }
    else
    {
      rules_list->next = vsf_sysutil_malloc(sizeof(ip_rule));
      vsf_sysutil_fillbuff(rules_list->next, sizeof(ip_rule), 0);
      rules_list->next->prev = rules_list;
      rules_list = rules_list->next;
    }
    rules_list->remote_ip = addr_remote;
    rules_list->remote_mask = addr_mask;
    rules_list->codepage = codepage;
    str_copy (&rules_list->p_user, &login);
    addr_remote = 0;
  }
  while (!str_isempty(&s_rhs_chunk_str));

  if (addr_remote)
  {
    vsf_sysutil_sockaddr_clear(&addr_remote);
    vsf_sysutil_sockaddr_clear(&addr_mask);
  }

  str_free(&s_lhs_chunk_str);
  str_free(&s_rhs_chunk_str);
  str_free(&login);
  str_free(&s_ip);
}

